<?php

namespace Drupal\commerce_vado\EventSubscriber;

use Drupal\commerce_cart\Event\CartEntityAddEvent;
use Drupal\commerce_cart\Event\CartEvents;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\Entity\OrderItemInterface;
use Drupal\commerce_order\Event\OrderEvent;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\RounderInterface;
use Drupal\commerce_vado\Entity\VadoGroupItem;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\commerce_order\Event\OrderItemEvent;
use Drupal\commerce_order\Event\OrderEvents;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Responds to preSave event to add an addon ot an order.
 */
class VadoEventSubscriber implements EventSubscriberInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The price rounder.
   *
   * @var \Drupal\commerce_price\RounderInterface
   */
  protected $rounder;

  /**
   * The constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_price\RounderInterface $rounder
   *   The price rounder.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, RounderInterface $rounder) {
    $this->entityTypeManager = $entity_type_manager;
    $this->rounder = $rounder;
  }

  /**
   * Get subscribed events.
   *
   * @return array
   *   The subscribed events.
   */
  public static function getSubscribedEvents() {
    $events = [
      CartEvents::CART_ENTITY_ADD => ['cartEntityAddEvent', -99],
      OrderEvents::ORDER_PRESAVE => ['orderPreSaveEvent', -99],
      OrderEvents::ORDER_ITEM_DELETE => ['orderItemDeleteEvent', -99],
    ];
    return $events;
  }

  /**
   * Adds addon to the cart if it's not already there, or locks the quantity.
   */
  public function cartEntityAddEvent(CartEntityAddEvent $event) {
    $order_item = $event->getOrderItem();
    $order = $event->getCart();
    $order_item_storage = $this->entityTypeManager->getStorage('commerce_order_item');

    /** @var \Drupal\commerce_product\Entity\ProductVariationInterface $parent_variation */
    $parent_variation = $order_item->getPurchasedEntity();

    $sync_quantity = $parent_variation->hasField('sync_quantity');
    if ($sync_quantity && !$parent_variation->get('sync_quantity')->isEmpty()) {
      $sync_quantity = $parent_variation->get('sync_quantity')->value;
    }

    $child_order_items = $order_item->getData('commerce_vado_child_order_items', []);
    // only create addon items if sync quantity is OFF or the order item doesn't already have children.
    if (!$sync_quantity || empty($child_order_items)) {

      //create a discount_multiplier variable for field_variation_addon_multiplier field
      $variation_discount_multiplier = 1;
      if ($parent_variation->hasField('discount_multiplier') && !$parent_variation->get('discount_multiplier')->isEmpty()) {
        $variation_discount_multiplier = $parent_variation->get('discount_multiplier')->value;
      }

      $addon_data = [];
      if ($parent_variation->hasField('child_variations') && !$parent_variation->get('child_variations')->isEmpty()) {
        foreach ($parent_variation->get('child_variations')->referencedEntities() as $addon_variation) {
          $addon_data[] = [
            'variation' => $addon_variation,
            'discount_multiplier' => $variation_discount_multiplier,
          ];
        }
      }

      $selected_group_items = $order_item->getData('selected_addon_group_items', FALSE);
      if ($selected_group_items) {
        /** @var VadoGroupItem[] $addon_group_items */
        $addon_group_items = VadoGroupItem::loadMultiple($selected_group_items);
        if (!empty($addon_group_items)) {
          foreach ($addon_group_items as $addon_group_item) {
            $discount_multiplier = $variation_discount_multiplier;
            $group = $addon_group_item->getGroup();
            // Use the discount percentage from the group if available.
            // @TODO create discount percentage methods for group.
            if (!$group->get('discount_multiplier')->isEmpty()) {
              $discount_multiplier = $group->get('discount_multiplier')->value;
            }
            $addon_data[] = [
              'variation' => $addon_group_item->get('variation')->entity,
              'discount_multiplier' => $discount_multiplier,
            ];
          }
        }
      }

      //set the variable for the parent discount field
      $parent_discount = $parent_variation->hasField('parent_discount');
      if ($parent_discount && !$parent_variation->get('parent_discount')->isEmpty()) {
        $parent_discount = $parent_variation->get('parent_discount')->value;
      }
      
      //sets price resolver order item price back to it's original price.
      $order_item->setUnitPrice($parent_variation->getPrice()->multiply($parent_discount ? $variation_discount_multiplier : 1), TRUE);

      /** @var \Drupal\commerce_cart\OrderItemMatcherInterface $order_item_matcher */
      $order_item_matcher = \Drupal::service('commerce_cart.order_item_matcher');

      // The VADO OrderItemMatcher uses this ID to determine whether order items should combine.
      $vado_combo_id = $this->getVadoComboId($order_item, $addon_data);
      foreach ($addon_data as $data) {
        $child_variation = $data['variation'];
        $child_price = $child_variation->getPrice()->multiply($data['discount_multiplier']);
        $child_price = $this->rounder->round($child_price);
        $child_order_item = $order_item_storage->create([
          'title' => $child_variation->getOrderItemTitle(),
          'type' => $child_variation->getOrderItemTypeId(),
          'unit_price' => $child_price,
          'purchased_entity' => $child_variation->id(),
          'quantity' => $order_item->getQuantity(),
          'overridden_unit_price' => TRUE,
        ]);
        $child_order_item->setData('commerce_vado_synced_child_order_item', $sync_quantity);
        $child_order_item->setData('commerce_vado_parent_order_item', $order_item->id());
        $child_order_item->setData('commerce_vado_combo_id', $vado_combo_id);

        $matching_order_item = $order_item_matcher->match($child_order_item, $order->getItems());
        if ($matching_order_item) {
          $new_quantity = Calculator::add($matching_order_item->getQuantity(), $child_order_item->getQuantity());
          $matching_order_item->setQuantity($new_quantity);
          $matching_order_item->save();
        } else {
          $child_order_item->save();
          $child_order_items[] = $child_order_item->id();
          $order->addItem($child_order_item);
        }
      }

      $order_item->setData('commerce_vado_combo_id', $vado_combo_id);
      $matching_order_item = $order_item_matcher->match($order_item, $order->getItems());
      if ($matching_order_item) {
        $new_quantity = Calculator::add($matching_order_item->getQuantity(), $order_item->getQuantity());
        $matching_order_item->setQuantity($new_quantity);
        $matching_order_item->save();
        $order_item->delete();
      } else {
        $order_item->setData('commerce_vado_child_order_items', $child_order_items);
        $order_item->save();
      }
    }
  }

  /**
   * Gets a unique string that is the combination of the parent variation
   * ID and all of the addons variations ids that were selected it, separated by
   * a dash.
   *
   * Example:
   *  Parent Variation ID = 1
   *  Addon Variation IDs = 2, 4, 6, 7
   *
   * Output: 1-2-4-6-7
   *
   * @param \Drupal\commerce_order\Entity\OrderItemInterface $order_item
   *   The order item responsible for generating addons.
   * @param array $addon_data
   *   The variation addon data with the following keys:
   *     'variation' => ProductVariationInterface,
   *     'discount_multiplier' => string
   * @return string
   *   The vado combination ID.
   */
  protected function getVadoComboId(OrderItemInterface $order_item, array $addon_data) {
    $combo_id = $order_item->getPurchasedEntityId();
    $addon_variation_ids = array_map(function($data) {
      return $data['variation']->id();
    }, $addon_data);
    $combo_id .= implode("-", $addon_variation_ids);

    return $combo_id;
  }

  public function orderPreSaveEvent(OrderEvent $event) {
    $order_item_storage = $this->entityTypeManager->getStorage('commerce_order_item');
    $order = $event->getOrder();

    $needs_refresh = FALSE;
    foreach ($order->getItems() as $order_item) {
      $child_order_item_ids = $order_item->getData('commerce_vado_child_order_items', []);
      $child_order_items = $order_item_storage->loadMultiple($child_order_item_ids);
      foreach ($child_order_items as $child_order_item) {
        if ($child_order_item->getData('commerce_vado_synced_child_order_item')) {
          $quantity = $order_item->getQuantity();
          $child_order_item->setQuantity($quantity);
          $child_order_item->save();
          $needs_refresh = TRUE;
        }
      }
    }
    if ($needs_refresh) {
      $order->setRefreshState(OrderInterface::REFRESH_ON_LOAD);
    }
  }

  public function orderItemDeleteEvent(OrderItemEvent $event) {
    $order_item_storage = $this->entityTypeManager->getStorage('commerce_order_item');
    $order_item = $event->getOrderItem();
    $parent_variation = $order_item->getPurchasedEntity();
    $sync_quantity = $parent_variation->hasField('sync_quantity');
    if ($sync_quantity && !$parent_variation->get('sync_quantity')->isEmpty()) {
      $sync_quantity = $parent_variation->get('sync_quantity')->value;
    }
    if ($sync_quantity) {
      $child_order_item_ids = $order_item->getData('commerce_vado_child_order_items', []);
      if(!empty($child_order_item_ids)) {
        $child_order_items = $order_item_storage->loadMultiple($child_order_item_ids);
        $order_item_storage->delete($child_order_items);
      }
    }
  }
}
